94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
1174 lines
33 KiB
Python
1174 lines
33 KiB
Python
# SPDX-License-Identifier: Apache-2.0
|
|
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
|
|
# Adapted from https://github.com/vllm-project/vllm/blob/0384aa7150c4c9778efca041ffd1beb3ad2bd694/vllm/model_executor/layers/fla/ops/kda.py
|
|
# This file contains code copied from the flash-linear-attention project.
|
|
# The original source code was licensed under the MIT license and included
|
|
# the following copyright notice:
|
|
# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
|
|
|
|
from typing import Optional
|
|
|
|
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.srt.layers.attention.fla.chunk_delta_h import chunk_gated_delta_rule_fwd_h
|
|
from sglang.srt.layers.attention.fla.chunk_intra import chunk_kda_fwd_intra
|
|
from sglang.srt.layers.attention.fla.cumsum import chunk_local_cumsum
|
|
from sglang.srt.layers.attention.fla.fused_norm_gate import layer_norm_gated_fwd
|
|
from sglang.srt.layers.attention.fla.fused_recurrent import (
|
|
fused_recurrent_gated_delta_rule_fwd_kernel,
|
|
)
|
|
from sglang.srt.layers.attention.fla.index import (
|
|
prepare_chunk_indices,
|
|
)
|
|
from sglang.srt.layers.attention.fla.l2norm import l2norm_fwd
|
|
from sglang.srt.layers.attention.fla.op import exp, log
|
|
from sglang.srt.layers.attention.fla.utils import (
|
|
check_shared_mem,
|
|
is_intel,
|
|
)
|
|
|
|
if is_intel:
|
|
from sglang.srt.hardware_backend.xpu.kernels.fla.chunk_delta_h import (
|
|
chunk_gated_delta_rule_fwd_h,
|
|
)
|
|
|
|
|
|
BS_LIST = [32, 64] if check_shared_mem() else [16, 32]
|
|
|
|
|
|
def cdiv(a: int, b: int) -> int:
|
|
"""Ceiling division."""
|
|
return -(a // -b)
|
|
|
|
|
|
def next_power_of_2(n: int) -> int:
|
|
"""The next power of 2 (inclusive)"""
|
|
if n < 1:
|
|
return 1
|
|
return 1 << (n - 1).bit_length()
|
|
|
|
|
|
def fused_recurrent_kda_fwd(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
inplace_final_state: bool = True,
|
|
cu_seqlens: torch.LongTensor | None = None,
|
|
# ssm_state_indices: torch.Tensor | None = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
B, T, H, K, V = *k.shape, v.shape[-1]
|
|
HV = v.shape[2]
|
|
N = B if cu_seqlens is None else len(cu_seqlens) - 1
|
|
BK, BV = next_power_of_2(K), min(next_power_of_2(V), 8)
|
|
NK, NV = cdiv(K, BK), cdiv(V, BV)
|
|
assert NK == 1, "NK > 1 is not supported yet"
|
|
num_stages = 3
|
|
num_warps = 1
|
|
|
|
o = q.new_empty(NK, *v.shape)
|
|
if inplace_final_state:
|
|
final_state = initial_state
|
|
else:
|
|
final_state = q.new_empty(N, HV, V, K, dtype=initial_state.dtype)
|
|
|
|
stride_init_state_token = initial_state.stride(0)
|
|
stride_final_state_token = final_state.stride(0)
|
|
|
|
# if ssm_state_indices is None:
|
|
# stride_indices_seq, stride_indices_tok = 1, 1
|
|
# elif ssm_state_indices.ndim == 1:
|
|
# stride_indices_seq, stride_indices_tok = ssm_state_indices.stride(0), 1
|
|
# else:
|
|
# stride_indices_seq, stride_indices_tok = ssm_state_indices.stride()
|
|
|
|
grid = (NK, NV, N * HV)
|
|
fused_recurrent_gated_delta_rule_fwd_kernel[grid](
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
g=g,
|
|
beta=beta,
|
|
o=o,
|
|
h0=initial_state,
|
|
ht=final_state,
|
|
cu_seqlens=cu_seqlens,
|
|
# ssm_state_indices=ssm_state_indices,
|
|
scale=scale,
|
|
# N=N,
|
|
T=T,
|
|
B=B,
|
|
H=H,
|
|
HV=HV,
|
|
K=K,
|
|
V=V,
|
|
BK=BK,
|
|
BV=BV,
|
|
# stride_init_state_token=stride_init_state_token,
|
|
# stride_final_state_token=stride_final_state_token,
|
|
# stride_indices_seq=stride_indices_seq,
|
|
# stride_indices_tok=stride_indices_tok,
|
|
USE_INITIAL_STATE=initial_state is not None,
|
|
STORE_FINAL_STATE=final_state is not None,
|
|
IS_BETA_HEADWISE=beta.ndim == v.ndim,
|
|
USE_QK_L2NORM_IN_KERNEL=use_qk_l2norm_in_kernel,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
# INPLACE_FINAL_STATE=inplace_final_state,
|
|
IS_KDA=True,
|
|
num_warps=num_warps,
|
|
num_stages=num_stages,
|
|
)
|
|
|
|
o = o.squeeze(0)
|
|
return o, final_state
|
|
|
|
|
|
def fused_recurrent_kda(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor = None,
|
|
scale: float = None,
|
|
initial_state: torch.Tensor = None,
|
|
inplace_final_state: bool = True,
|
|
use_qk_l2norm_in_kernel: bool = True,
|
|
cu_seqlens: torch.LongTensor | None = None,
|
|
# ssm_state_indices: torch.LongTensor | None = None,
|
|
**kwargs,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
if cu_seqlens is not None and q.shape[0] != 1:
|
|
raise ValueError(
|
|
f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
|
|
f"Please flatten variable-length inputs before processing."
|
|
)
|
|
if scale is None:
|
|
scale = k.shape[-1] ** -0.5
|
|
|
|
o, final_state = fused_recurrent_kda_fwd(
|
|
q=q.contiguous(),
|
|
k=k.contiguous(),
|
|
v=v.contiguous(),
|
|
g=g.contiguous(),
|
|
beta=beta.contiguous(),
|
|
scale=scale,
|
|
initial_state=initial_state,
|
|
inplace_final_state=inplace_final_state,
|
|
cu_seqlens=cu_seqlens,
|
|
# ssm_state_indices=ssm_state_indices,
|
|
use_qk_l2norm_in_kernel=use_qk_l2norm_in_kernel,
|
|
)
|
|
return o, final_state
|
|
|
|
|
|
def rms_norm_gated(
|
|
x: torch.Tensor,
|
|
g: torch.Tensor,
|
|
weight: torch.Tensor,
|
|
bias: torch.Tensor,
|
|
activation: str = "swish",
|
|
residual: torch.Tensor | None = None,
|
|
prenorm: bool = False,
|
|
residual_in_fp32: bool = False,
|
|
eps: float = 1e-6,
|
|
):
|
|
x_shape_og = x.shape
|
|
# reshape input data into 2D tensor
|
|
x = x.contiguous().reshape(-1, x.shape[-1])
|
|
g = g.contiguous().reshape(-1, g.shape[-1])
|
|
if residual is not None:
|
|
assert residual.shape == x_shape_og
|
|
residual = residual.contiguous().reshape(-1, residual.shape[-1])
|
|
residual_dtype = (
|
|
residual.dtype
|
|
if residual is not None
|
|
else (torch.float if residual_in_fp32 else None)
|
|
)
|
|
y, _, _, residual_out = layer_norm_gated_fwd(
|
|
x=x,
|
|
g=g,
|
|
weight=weight,
|
|
bias=bias,
|
|
activation=activation,
|
|
eps=eps,
|
|
residual=residual,
|
|
residual_dtype=residual_dtype,
|
|
is_rms_norm=True,
|
|
)
|
|
y = y.reshape(x_shape_og)
|
|
return y if not prenorm else (y, residual_out.reshape(x_shape_og))
|
|
|
|
|
|
@triton.autotune(
|
|
configs=[
|
|
triton.Config({"BK": BK}, num_warps=num_warps, num_stages=num_stages)
|
|
for BK in [32, 64]
|
|
for num_warps in [1, 2, 4, 8]
|
|
for num_stages in [2, 3, 4]
|
|
],
|
|
key=["BC", "IS_VARLEN"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_kda_scaled_dot_kkt_fwd_kernel_intra_sub_inter(
|
|
q,
|
|
k,
|
|
g,
|
|
beta,
|
|
A,
|
|
Aqk,
|
|
scale,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
T,
|
|
H: tl.constexpr,
|
|
K: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BC: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
NC: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
):
|
|
i_t, i_c, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
i_i, i_j = i_c // NC, i_c % NC
|
|
if IS_VARLEN:
|
|
i_n, i_t = (
|
|
tl.load(chunk_indices + i_t * 2).to(tl.int32),
|
|
tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32),
|
|
)
|
|
bos, eos = (
|
|
tl.load(cu_seqlens + i_n).to(tl.int32),
|
|
tl.load(cu_seqlens + i_n + 1).to(tl.int32),
|
|
)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
if i_t * BT + i_i * BC >= T:
|
|
return
|
|
if i_i <= i_j:
|
|
return
|
|
|
|
q += (bos * H + i_h) * K
|
|
k += (bos * H + i_h) * K
|
|
g += (bos * H + i_h) * K
|
|
A += (bos * H + i_h) * BT
|
|
Aqk += (bos * H + i_h) * BT
|
|
|
|
p_b = tl.make_block_ptr(
|
|
beta + bos * H + i_h, (T,), (H,), (i_t * BT + i_i * BC,), (BC,), (0,)
|
|
)
|
|
b_b = tl.load(p_b, boundary_check=(0,))
|
|
|
|
b_A = tl.zeros([BC, BC], dtype=tl.float32)
|
|
b_Aqk = tl.zeros([BC, BC], dtype=tl.float32)
|
|
for i_k in range(tl.cdiv(K, BK)):
|
|
p_q = tl.make_block_ptr(
|
|
q, (T, K), (H * K, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0)
|
|
)
|
|
p_k = tl.make_block_ptr(
|
|
k, (T, K), (H * K, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0)
|
|
)
|
|
p_g = tl.make_block_ptr(
|
|
g, (T, K), (H * K, 1), (i_t * BT + i_i * BC, i_k * BK), (BC, BK), (1, 0)
|
|
)
|
|
b_kt = tl.make_block_ptr(
|
|
k, (K, T), (1, H * K), (i_k * BK, i_t * BT + i_j * BC), (BK, BC), (0, 1)
|
|
)
|
|
p_gk = tl.make_block_ptr(
|
|
g, (K, T), (1, H * K), (i_k * BK, i_t * BT + i_j * BC), (BK, BC), (0, 1)
|
|
)
|
|
|
|
o_k = i_k * BK + tl.arange(0, BK)
|
|
m_k = o_k < K
|
|
# [BK,]
|
|
b_gn = tl.load(g + (i_t * BT + i_i * BC) * H * K + o_k, mask=m_k, other=0)
|
|
# [BC, BK]
|
|
b_g = tl.load(p_g, boundary_check=(0, 1))
|
|
b_k = tl.load(p_k, boundary_check=(0, 1)) * exp(b_g - b_gn[None, :])
|
|
# [BK, BC]
|
|
b_gk = tl.load(p_gk, boundary_check=(0, 1))
|
|
b_kt = tl.load(b_kt, boundary_check=(0, 1))
|
|
# [BC, BC]
|
|
b_ktg = b_kt * exp(b_gn[:, None] - b_gk)
|
|
b_A += tl.dot(b_k, b_ktg)
|
|
|
|
b_q = tl.load(p_q, boundary_check=(0, 1))
|
|
b_qg = b_q * exp(b_g - b_gn[None, :]) * scale
|
|
b_Aqk += tl.dot(b_qg, b_ktg)
|
|
|
|
b_A *= b_b[:, None]
|
|
|
|
p_A = tl.make_block_ptr(
|
|
A, (T, BT), (H * BT, 1), (i_t * BT + i_i * BC, i_j * BC), (BC, BC), (1, 0)
|
|
)
|
|
tl.store(p_A, b_A.to(A.dtype.element_ty), boundary_check=(0, 1))
|
|
p_Aqk = tl.make_block_ptr(
|
|
Aqk, (T, BT), (H * BT, 1), (i_t * BT + i_i * BC, i_j * BC), (BC, BC), (1, 0)
|
|
)
|
|
tl.store(p_Aqk, b_Aqk.to(Aqk.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
@triton.autotune(
|
|
configs=[triton.Config({}, num_warps=num_warps) for num_warps in [1, 2, 4, 8]],
|
|
key=["BK", "BT", "IS_VARLEN"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_kda_scaled_dot_kkt_fwd_kernel_intra_sub_intra(
|
|
q,
|
|
k,
|
|
g,
|
|
beta,
|
|
A,
|
|
Aqk,
|
|
scale,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
T,
|
|
H: tl.constexpr,
|
|
K: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BC: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
):
|
|
i_t, i_i, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_n, i_t = (
|
|
tl.load(chunk_indices + i_t * 2).to(tl.int32),
|
|
tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32),
|
|
)
|
|
bos, eos = (
|
|
tl.load(cu_seqlens + i_n).to(tl.int32),
|
|
tl.load(cu_seqlens + i_n + 1).to(tl.int32),
|
|
)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
if i_t * BT + i_i * BC >= T:
|
|
return
|
|
|
|
o_i = tl.arange(0, BC)
|
|
o_k = tl.arange(0, BK)
|
|
m_k = o_k < K
|
|
m_A = (i_t * BT + i_i * BC + o_i) < T
|
|
o_A = (bos + i_t * BT + i_i * BC + o_i) * H * BT + i_h * BT + i_i * BC
|
|
|
|
p_q = tl.make_block_ptr(
|
|
q + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT + i_i * BC, 0),
|
|
(BC, BK),
|
|
(1, 0),
|
|
)
|
|
p_k = tl.make_block_ptr(
|
|
k + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT + i_i * BC, 0),
|
|
(BC, BK),
|
|
(1, 0),
|
|
)
|
|
p_g = tl.make_block_ptr(
|
|
g + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT + i_i * BC, 0),
|
|
(BC, BK),
|
|
(1, 0),
|
|
)
|
|
b_q = tl.load(p_q, boundary_check=(0, 1))
|
|
b_k = tl.load(p_k, boundary_check=(0, 1))
|
|
b_g = tl.load(p_g, boundary_check=(0, 1))
|
|
|
|
p_b = beta + (bos + i_t * BT + i_i * BC + o_i) * H + i_h
|
|
b_k = b_k * tl.load(p_b, mask=m_A, other=0)[:, None]
|
|
|
|
p_kt = k + (bos + i_t * BT + i_i * BC) * H * K + i_h * K + o_k
|
|
p_gk = g + (bos + i_t * BT + i_i * BC) * H * K + i_h * K + o_k
|
|
|
|
for j in range(0, min(BC, T - i_t * BT - i_i * BC)):
|
|
b_kt = tl.load(p_kt, mask=m_k, other=0).to(tl.float32)
|
|
b_gk = tl.load(p_gk, mask=m_k, other=0).to(tl.float32)
|
|
b_ktg = b_kt[None, :] * exp(b_g - b_gk[None, :])
|
|
b_A = tl.sum(b_k * b_ktg, 1)
|
|
b_A = tl.where(o_i > j, b_A, 0.0)
|
|
b_Aqk = tl.sum(b_q * b_ktg, 1)
|
|
b_Aqk = tl.where(o_i >= j, b_Aqk * scale, 0.0)
|
|
tl.store(A + o_A + j, b_A, mask=m_A)
|
|
tl.store(Aqk + o_A + j, b_Aqk, mask=m_A)
|
|
p_kt += H * K
|
|
p_gk += H * K
|
|
|
|
|
|
def chunk_kda_scaled_dot_kkt_fwd(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
gk: torch.Tensor | None = None,
|
|
beta: torch.Tensor | None = None,
|
|
scale: float | None = None,
|
|
cu_seqlens: torch.LongTensor | None = None,
|
|
chunk_size: int = 64,
|
|
output_dtype: torch.dtype = torch.float32,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
r"""
|
|
Compute beta * K * K^T.
|
|
|
|
Args:
|
|
k (torch.Tensor):
|
|
The key tensor of shape `[B, T, H, K]`.
|
|
beta (torch.Tensor):
|
|
The beta tensor of shape `[B, T, H]`.
|
|
gk (torch.Tensor):
|
|
The cumulative sum of the gate tensor of shape `[B, T, H, K]` applied to the key tensor. Default: `None`.
|
|
cu_seqlens (torch.LongTensor):
|
|
The cumulative sequence lengths of the input tensor.
|
|
Default: None
|
|
chunk_size (int):
|
|
The chunk size. Default: 64.
|
|
output_dtype (torch.dtype):
|
|
The dtype of the output tensor. Default: `torch.float32`
|
|
|
|
Returns:
|
|
beta * K * K^T of shape `[B, T, H, BT]` where `BT` is the chunk size.
|
|
"""
|
|
B, T, H, K = k.shape
|
|
assert K <= 256
|
|
BT = chunk_size
|
|
chunk_indices = (
|
|
prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
|
|
)
|
|
NT = cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
|
|
BC = min(16, BT)
|
|
NC = cdiv(BT, BC)
|
|
BK = max(next_power_of_2(K), 16)
|
|
A = torch.zeros(B, T, H, BT, device=k.device, dtype=output_dtype)
|
|
Aqk = torch.zeros(B, T, H, BT, device=k.device, dtype=output_dtype)
|
|
grid = (NT, NC * NC, B * H)
|
|
chunk_kda_scaled_dot_kkt_fwd_kernel_intra_sub_inter[grid](
|
|
q=q,
|
|
k=k,
|
|
g=gk,
|
|
beta=beta,
|
|
A=A,
|
|
Aqk=Aqk,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
T=T,
|
|
H=H,
|
|
K=K,
|
|
BT=BT,
|
|
BC=BC,
|
|
NC=NC,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
)
|
|
|
|
grid = (NT, NC, B * H)
|
|
chunk_kda_scaled_dot_kkt_fwd_kernel_intra_sub_intra[grid](
|
|
q=q,
|
|
k=k,
|
|
g=gk,
|
|
beta=beta,
|
|
A=A,
|
|
Aqk=Aqk,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
T=T,
|
|
H=H,
|
|
K=K,
|
|
BT=BT,
|
|
BC=BC,
|
|
BK=BK,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
)
|
|
return A, Aqk
|
|
|
|
|
|
@triton.autotune(
|
|
configs=[
|
|
triton.Config({"BK": BK, "BV": BV}, num_warps=num_warps, num_stages=num_stages)
|
|
for BK in [64, 128]
|
|
for BV in [64, 128]
|
|
for num_warps in [2, 4, 8]
|
|
for num_stages in [2, 3, 4]
|
|
],
|
|
key=["H", "K", "V", "BT", "IS_VARLEN"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def recompute_w_u_fwd_kernel(
|
|
q,
|
|
k,
|
|
qg,
|
|
kg,
|
|
v,
|
|
beta,
|
|
w,
|
|
u,
|
|
A,
|
|
gk,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
T,
|
|
H: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
STORE_QG: tl.constexpr,
|
|
STORE_KG: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
DOT_PRECISION: tl.constexpr,
|
|
):
|
|
i_t, i_bh = tl.program_id(0), tl.program_id(1)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_n, i_t = (
|
|
tl.load(chunk_indices + i_t * 2).to(tl.int32),
|
|
tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32),
|
|
)
|
|
bos, eos = (
|
|
tl.load(cu_seqlens + i_n).to(tl.int32),
|
|
tl.load(cu_seqlens + i_n + 1).to(tl.int32),
|
|
)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
p_b = tl.make_block_ptr(beta + bos * H + i_h, (T,), (H,), (i_t * BT,), (BT,), (0,))
|
|
b_b = tl.load(p_b, boundary_check=(0,))
|
|
|
|
p_A = tl.make_block_ptr(
|
|
A + (bos * H + i_h) * BT, (T, BT), (H * BT, 1), (i_t * BT, 0), (BT, BT), (1, 0)
|
|
)
|
|
b_A = tl.load(p_A, boundary_check=(0, 1))
|
|
|
|
for i_v in range(tl.cdiv(V, BV)):
|
|
p_v = tl.make_block_ptr(
|
|
v + (bos * H + i_h) * V,
|
|
(T, V),
|
|
(H * V, 1),
|
|
(i_t * BT, i_v * BV),
|
|
(BT, BV),
|
|
(1, 0),
|
|
)
|
|
p_u = tl.make_block_ptr(
|
|
u + (bos * H + i_h) * V,
|
|
(T, V),
|
|
(H * V, 1),
|
|
(i_t * BT, i_v * BV),
|
|
(BT, BV),
|
|
(1, 0),
|
|
)
|
|
b_v = tl.load(p_v, boundary_check=(0, 1))
|
|
b_vb = (b_v * b_b[:, None]).to(b_v.dtype)
|
|
b_u = tl.dot(b_A, b_vb, input_precision=DOT_PRECISION)
|
|
tl.store(p_u, b_u.to(p_u.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
for i_k in range(tl.cdiv(K, BK)):
|
|
p_w = tl.make_block_ptr(
|
|
w + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
p_k = tl.make_block_ptr(
|
|
k + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
b_k = tl.load(p_k, boundary_check=(0, 1))
|
|
b_kb = b_k * b_b[:, None]
|
|
|
|
p_gk = tl.make_block_ptr(
|
|
gk + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
b_gk = tl.load(p_gk, boundary_check=(0, 1))
|
|
b_kb *= exp(b_gk)
|
|
if STORE_QG:
|
|
p_q = tl.make_block_ptr(
|
|
q + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
p_qg = tl.make_block_ptr(
|
|
qg + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
b_q = tl.load(p_q, boundary_check=(0, 1))
|
|
b_qg = b_q * exp(b_gk)
|
|
tl.store(p_qg, b_qg.to(p_qg.dtype.element_ty), boundary_check=(0, 1))
|
|
if STORE_KG:
|
|
last_idx = min(i_t * BT + BT, T) - 1
|
|
|
|
o_k = i_k * BK + tl.arange(0, BK)
|
|
m_k = o_k < K
|
|
b_gn = tl.load(
|
|
gk + ((bos + last_idx) * H + i_h) * K + o_k, mask=m_k, other=0.0
|
|
)
|
|
b_kg = b_k * exp(b_gn - b_gk)
|
|
|
|
p_kg = tl.make_block_ptr(
|
|
kg + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
tl.store(p_kg, b_kg.to(p_kg.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
b_w = tl.dot(b_A, b_kb.to(b_k.dtype))
|
|
tl.store(p_w, b_w.to(p_w.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
def recompute_w_u_fwd(
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
A: torch.Tensor,
|
|
q: torch.Tensor | None = None,
|
|
gk: torch.Tensor | None = None,
|
|
cu_seqlens: torch.LongTensor | None = None,
|
|
chunk_indices: torch.LongTensor | None = None,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
B, T, H, K, V = *k.shape, v.shape[-1]
|
|
BT = A.shape[-1]
|
|
|
|
if chunk_indices is None and cu_seqlens is not None:
|
|
chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
|
|
NT = cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
|
|
w = torch.empty_like(k)
|
|
u = torch.empty_like(v)
|
|
kg = torch.empty_like(k) if gk is not None else None
|
|
recompute_w_u_fwd_kernel[(NT, B * H)](
|
|
q=q,
|
|
k=k,
|
|
qg=None,
|
|
kg=kg,
|
|
v=v,
|
|
beta=beta,
|
|
w=w,
|
|
u=u,
|
|
A=A,
|
|
gk=gk,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
T=T,
|
|
H=H,
|
|
K=K,
|
|
V=V,
|
|
BT=BT,
|
|
STORE_QG=False,
|
|
STORE_KG=kg is not None,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
DOT_PRECISION="tf32",
|
|
)
|
|
return w, u, None, kg
|
|
|
|
|
|
@triton.autotune(
|
|
configs=[
|
|
triton.Config({"BK": BK, "BV": BV}, num_warps=num_warps, num_stages=num_stages)
|
|
for BK in [64]
|
|
for BV in [64]
|
|
for num_warps in [2, 4, 8]
|
|
for num_stages in [2, 3, 4]
|
|
],
|
|
key=["BT", "IS_VARLEN"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def chunk_gla_fwd_kernel_o(
|
|
q,
|
|
v,
|
|
g,
|
|
h,
|
|
o,
|
|
A,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
scale,
|
|
T,
|
|
H: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
):
|
|
i_v, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_tg = i_t
|
|
i_n, i_t = (
|
|
tl.load(chunk_indices + i_t * 2).to(tl.int32),
|
|
tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32),
|
|
)
|
|
bos, eos = (
|
|
tl.load(cu_seqlens + i_n).to(tl.int32),
|
|
tl.load(cu_seqlens + i_n + 1).to(tl.int32),
|
|
)
|
|
T = eos - bos
|
|
NT = tl.cdiv(T, BT)
|
|
else:
|
|
NT = tl.cdiv(T, BT)
|
|
i_tg = i_b * NT + i_t
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
m_s = tl.arange(0, BT)[:, None] >= tl.arange(0, BT)[None, :]
|
|
|
|
b_o = tl.zeros([BT, BV], dtype=tl.float32)
|
|
for i_k in range(tl.cdiv(K, BK)):
|
|
p_q = tl.make_block_ptr(
|
|
q + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
p_g = tl.make_block_ptr(
|
|
g + (bos * H + i_h) * K,
|
|
(T, K),
|
|
(H * K, 1),
|
|
(i_t * BT, i_k * BK),
|
|
(BT, BK),
|
|
(1, 0),
|
|
)
|
|
p_h = tl.make_block_ptr(
|
|
h + (i_tg * H + i_h) * V * K,
|
|
(V, K),
|
|
(K, 1),
|
|
(i_v * BV, i_k * BK),
|
|
(BV, BK),
|
|
(1, 0),
|
|
)
|
|
|
|
# [BT, BK]
|
|
b_q = tl.load(p_q, boundary_check=(0, 1))
|
|
b_q = (b_q * scale).to(b_q.dtype)
|
|
# [BT, BK]
|
|
b_g = tl.load(p_g, boundary_check=(0, 1))
|
|
# [BT, BK]
|
|
b_qg = (b_q * exp(b_g)).to(b_q.dtype)
|
|
# [BK, BV]
|
|
b_h = tl.load(p_h, boundary_check=(0, 1))
|
|
# works but dkw, owing to divine benevolence
|
|
# [BT, BV]
|
|
if i_k >= 0:
|
|
b_o += tl.dot(b_qg, tl.trans(b_h).to(b_qg.dtype))
|
|
p_v = tl.make_block_ptr(
|
|
v + (bos * H + i_h) * V,
|
|
(T, V),
|
|
(H * V, 1),
|
|
(i_t * BT, i_v * BV),
|
|
(BT, BV),
|
|
(1, 0),
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o + (bos * H + i_h) * V,
|
|
(T, V),
|
|
(H * V, 1),
|
|
(i_t * BT, i_v * BV),
|
|
(BT, BV),
|
|
(1, 0),
|
|
)
|
|
p_A = tl.make_block_ptr(
|
|
A + (bos * H + i_h) * BT, (T, BT), (H * BT, 1), (i_t * BT, 0), (BT, BT), (1, 0)
|
|
)
|
|
# [BT, BV]
|
|
b_v = tl.load(p_v, boundary_check=(0, 1))
|
|
# [BT, BT]
|
|
b_A = tl.load(p_A, boundary_check=(0, 1))
|
|
b_A = tl.where(m_s, b_A, 0.0).to(b_v.dtype)
|
|
b_o += tl.dot(b_A, b_v)
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
def chunk_gla_fwd_o_gk(
|
|
q: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
A: torch.Tensor,
|
|
h: torch.Tensor,
|
|
o: torch.Tensor,
|
|
scale: float,
|
|
cu_seqlens: torch.LongTensor | None = None,
|
|
chunk_size: int = 64,
|
|
chunk_indices: torch.LongTensor | None = None,
|
|
):
|
|
B, T, H, K, V = *q.shape, v.shape[-1]
|
|
BT = chunk_size
|
|
|
|
if chunk_indices is None and cu_seqlens is not None:
|
|
chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
|
|
NT = cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
|
|
def grid(meta):
|
|
return (cdiv(V, meta["BV"]), NT, B * H)
|
|
|
|
chunk_gla_fwd_kernel_o[grid](
|
|
q=q,
|
|
v=v,
|
|
g=g,
|
|
h=h,
|
|
o=o,
|
|
A=A,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
scale=scale,
|
|
T=T,
|
|
H=H,
|
|
K=K,
|
|
V=V,
|
|
BT=BT,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
)
|
|
return o
|
|
|
|
|
|
@triton.jit
|
|
def softplus_fwd(x):
|
|
"""Standard softplus: log(1 + exp(x)), with linear approx for large x."""
|
|
return tl.where(x < 20.0, log(1.0 + exp(x)), x)
|
|
|
|
|
|
@triton.heuristics(
|
|
{
|
|
"HAS_BIAS": lambda args: args["dt_bias"] is not None,
|
|
"HAS_SCALE": lambda args: args["scale"] is not None,
|
|
"IS_VARLEN": lambda args: args["cu_seqlens"] is not None,
|
|
"USE_LOWER_BOUND": lambda args: args["lower_bound"] is not None,
|
|
}
|
|
)
|
|
@triton.autotune(
|
|
configs=[
|
|
triton.Config({"BS": BS}, num_warps=num_warps)
|
|
for BS in BS_LIST
|
|
for num_warps in [2, 4, 8]
|
|
],
|
|
key=["H", "S", "BT", "IS_VARLEN"],
|
|
)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def kda_gate_chunk_cumsum_vector_kernel(
|
|
s,
|
|
A_log,
|
|
dt_bias,
|
|
o,
|
|
scale,
|
|
cu_seqlens,
|
|
chunk_indices,
|
|
lower_bound,
|
|
T,
|
|
H: tl.constexpr,
|
|
S: tl.constexpr,
|
|
BT: tl.constexpr,
|
|
BS: tl.constexpr,
|
|
HAS_BIAS: tl.constexpr,
|
|
HAS_SCALE: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
USE_LOWER_BOUND: tl.constexpr,
|
|
):
|
|
i_s, i_t, i_bh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_b, i_h = i_bh // H, i_bh % H
|
|
if IS_VARLEN:
|
|
i_n, i_t = (
|
|
tl.load(chunk_indices + i_t * 2).to(tl.int32),
|
|
tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32),
|
|
)
|
|
bos, eos = (
|
|
tl.load(cu_seqlens + i_n).to(tl.int32),
|
|
tl.load(cu_seqlens + i_n + 1).to(tl.int32),
|
|
)
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_b * T, i_b * T + T
|
|
|
|
p_s = tl.make_block_ptr(
|
|
s + (bos * H + i_h) * S,
|
|
(T, S),
|
|
(H * S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
p_o = tl.make_block_ptr(
|
|
o + (bos * H + i_h) * S,
|
|
(T, S),
|
|
(H * S, 1),
|
|
(i_t * BT, i_s * BS),
|
|
(BT, BS),
|
|
(1, 0),
|
|
)
|
|
# [BT, BS]
|
|
b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
|
|
|
|
if HAS_BIAS:
|
|
p_b = tl.make_block_ptr(
|
|
dt_bias + i_h * S,
|
|
(S,),
|
|
(1,),
|
|
(i_s * BS,),
|
|
(BS,),
|
|
(0,),
|
|
)
|
|
b_bias = tl.load(p_b, boundary_check=(0,)).to(tl.float32)
|
|
b_s = b_s + b_bias[None, :]
|
|
|
|
b_A = tl.load(A_log + i_h).to(tl.float32)
|
|
if not USE_LOWER_BOUND:
|
|
# Standard gate: -exp(A_log) * softplus(g + bias)
|
|
b_gate = -exp(b_A) * softplus_fwd(b_s)
|
|
else:
|
|
# Safe gate: lower_bound * sigmoid(exp(A_log) * (g + bias))
|
|
b_gate = lower_bound * tl.sigmoid(exp(b_A) * b_s)
|
|
|
|
# Chunk-local cumulative sum
|
|
b_o = tl.cumsum(b_gate, axis=0)
|
|
|
|
if HAS_SCALE:
|
|
b_o *= scale
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), boundary_check=(0, 1))
|
|
|
|
|
|
def kda_gate_chunk_cumsum(
|
|
g: torch.Tensor,
|
|
A_log: torch.Tensor,
|
|
chunk_size: int,
|
|
scale: float = None,
|
|
dt_bias: Optional[torch.Tensor] = None,
|
|
cu_seqlens: Optional[torch.Tensor] = None,
|
|
output_dtype: Optional[torch.dtype] = torch.float,
|
|
chunk_indices: Optional[torch.LongTensor] = None,
|
|
lower_bound: Optional[float] = None,
|
|
) -> torch.Tensor:
|
|
"""
|
|
Fused KDA gate activation + chunk-local cumulative sum.
|
|
|
|
Combines two memory-bound kernels into one:
|
|
1. Gate activation: g = -exp(A_log) * softplus(raw_g + dt_bias)
|
|
2. Chunk-local cumsum along the time axis
|
|
|
|
Args:
|
|
g: Raw gate tensor of shape [B, T, H, K] (before activation).
|
|
A_log: Per-head log-scale parameter, [H] elements (any shape, numel=H).
|
|
chunk_size: Chunk size for cumsum (must be power of 2).
|
|
scale: Optional scale factor applied to output.
|
|
dt_bias: Optional per-head bias, flat [H*K] elements.
|
|
cu_seqlens: Cumulative sequence lengths for variable-length input.
|
|
output_dtype: Output dtype (default float32).
|
|
chunk_indices: Pre-computed chunk indices for varlen mode.
|
|
lower_bound: If set, use safe gate: lower_bound * sigmoid(exp(A_log) * g).
|
|
|
|
Returns:
|
|
Cumulative-summed gated tensor of shape [B, T, H, K].
|
|
"""
|
|
if cu_seqlens is not None:
|
|
assert (
|
|
g.shape[0] == 1
|
|
), "Only batch size 1 is supported when cu_seqlens are provided"
|
|
assert len(g.shape) == 4
|
|
B, T, H, S = g.shape
|
|
BT = chunk_size
|
|
if chunk_indices is None and cu_seqlens is not None:
|
|
chunk_indices = prepare_chunk_indices(cu_seqlens, BT)
|
|
NT = cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
|
|
assert chunk_size == 2 ** (
|
|
chunk_size.bit_length() - 1
|
|
), "chunk_size must be a power of 2"
|
|
|
|
g_org, g = g, torch.empty_like(g, dtype=output_dtype or g.dtype)
|
|
|
|
def grid(meta):
|
|
return (cdiv(meta["S"], meta["BS"]), NT, B * H)
|
|
|
|
kda_gate_chunk_cumsum_vector_kernel[grid](
|
|
s=g_org,
|
|
A_log=A_log,
|
|
dt_bias=dt_bias,
|
|
o=g,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
lower_bound=lower_bound,
|
|
T=T,
|
|
H=H,
|
|
S=S,
|
|
BT=BT,
|
|
)
|
|
return g
|
|
|
|
|
|
def chunk_kda_fwd(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
initial_state_indices: torch.Tensor,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
A_log: Optional[torch.Tensor] = None,
|
|
dt_bias: Optional[torch.Tensor] = None,
|
|
lower_bound: Optional[float] = None,
|
|
):
|
|
chunk_size = 64
|
|
# Pre-compute chunk indices once and thread through all downstream kernels.
|
|
# Without this, each of the 4 callees would recompute independently.
|
|
chunk_indices = (
|
|
prepare_chunk_indices(cu_seqlens, chunk_size)
|
|
if cu_seqlens is not None
|
|
else None
|
|
)
|
|
|
|
if A_log is not None:
|
|
# Fused: gate activation + chunk-local cumsum in one kernel.
|
|
# g is raw gate (before activation); A_log, dt_bias drive the activation.
|
|
g = kda_gate_chunk_cumsum(
|
|
g,
|
|
A_log=A_log,
|
|
chunk_size=chunk_size,
|
|
dt_bias=dt_bias,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
lower_bound=lower_bound,
|
|
)
|
|
else:
|
|
# g is already gate-activated by caller; just do cumsum.
|
|
g = chunk_local_cumsum(
|
|
g,
|
|
chunk_size=chunk_size,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
)
|
|
|
|
# FUSE_DIAGONAL (fold diagonal-block compute into inter+solve) and
|
|
# FUSE_RECOMPUTE (also fold w/u/kg recompute) save kernel launches and HBM
|
|
# round-trips, but cost register footprint per CTA. Wins at small grid
|
|
# where launch overhead dominates; loses at large grid where the extra
|
|
# register pressure spills. Gate both on the same grid heuristic.
|
|
# Total CTAs in inter_solve_fused = NT * B * H_per_rank. For varlen,
|
|
# chunks don't cross sequence boundaries, so per-sequence ceil-divs sum to
|
|
# more than cdiv(total_tokens, chunk_size); use chunk_indices.shape[0] which
|
|
# already enumerates all (seq, chunk) pairs.
|
|
_NT_pr = (
|
|
triton.cdiv(q.shape[1], chunk_size)
|
|
if cu_seqlens is None
|
|
else chunk_indices.shape[0]
|
|
)
|
|
_H_pr = q.shape[-2]
|
|
_B = q.shape[0]
|
|
_small_grid = _B * _NT_pr * _H_pr <= 256
|
|
w, u, _, kg, Aqk, _ = chunk_kda_fwd_intra(
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
gk=g,
|
|
beta=beta,
|
|
scale=scale,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_size=chunk_size,
|
|
chunk_indices=chunk_indices,
|
|
fuse_diagonal=_small_grid,
|
|
fuse_recompute=_small_grid,
|
|
)
|
|
|
|
h, v_new = chunk_gated_delta_rule_fwd_h(
|
|
k=kg,
|
|
w=w,
|
|
u=u,
|
|
gk=g,
|
|
initial_state=initial_state,
|
|
initial_state_indices=initial_state_indices,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
)
|
|
del w, u, kg
|
|
|
|
o = chunk_gla_fwd_o_gk(
|
|
q=q,
|
|
v=v_new,
|
|
g=g,
|
|
A=Aqk,
|
|
h=h,
|
|
o=v,
|
|
scale=scale,
|
|
chunk_size=chunk_size,
|
|
cu_seqlens=cu_seqlens,
|
|
chunk_indices=chunk_indices,
|
|
)
|
|
del Aqk, v_new, h
|
|
|
|
return o
|
|
|
|
|
|
def chunk_kda(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float = None,
|
|
initial_state: torch.Tensor = None,
|
|
initial_state_indices: torch.Tensor = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
A_log: Optional[torch.Tensor] = None,
|
|
dt_bias: Optional[torch.Tensor] = None,
|
|
lower_bound: Optional[float] = None,
|
|
**kwargs,
|
|
):
|
|
if scale is None:
|
|
scale = k.shape[-1] ** -0.5
|
|
|
|
if use_qk_l2norm_in_kernel:
|
|
q = l2norm_fwd(q.contiguous())
|
|
k = l2norm_fwd(k.contiguous())
|
|
|
|
o = chunk_kda_fwd(
|
|
q=q,
|
|
k=k,
|
|
v=v.contiguous(),
|
|
g=g.contiguous(),
|
|
beta=beta.contiguous(),
|
|
scale=scale,
|
|
initial_state=initial_state,
|
|
initial_state_indices=initial_state_indices,
|
|
cu_seqlens=cu_seqlens,
|
|
A_log=A_log,
|
|
dt_bias=dt_bias,
|
|
lower_bound=lower_bound,
|
|
)
|
|
return o
|